@@ -73,14 +73,13 @@
 #include "compiler.h" /* write_mem_barrier */
 
 #include "rhd.h"
+#include "rhd_cs.h"
 #include "r5xx_accel.h"
 #include "r5xx_regs.h"
 
 extern struct R5xxRop R5xxRops[];
 
 struct R5xxXaaPrivate {
-    int scrnIndex;
-
     CARD32 dst_pitch_offset; /* Base value for R5XX_DST_PITCH_OFFSET */
     CARD32 control; /* Base value for R5XX_DP_GUI_MASTER_CNTL */
     CARD32 control_saved;
@@ -93,9 +92,9 @@ struct R5xxXaaPrivate {
     int scanline_x;
     int scanline_y;
     int scanline_w;
-    CARD32 scanline_h; /* R5xx */
-    CARD32 scanline_words; /* R5xx */
-    int scanline_bpp; /* Only used for ImageWrite -- R5xx */
+    CARD32 scanline_h;
+    CARD32 scanline_words;
+    int scanline_bpp; /* Only used for ImageWrite */
     int scanline_fg;
     int scanline_bg;
     int scanline_hpass;
@@ -106,6 +105,19 @@ struct R5xxXaaPrivate {
     CARD8 *BufferHook[1];
 };
 
+/*
+ *
+ */
+static void
+R5xxXAAIdle(ScrnInfoPtr pScrn)
+{
+    struct RhdCS *CS = RHDPTR(pScrn)->CS;
+
+    RHDCSFlush(CS);
+    RHDCSIdle(CS);
+    R5xx2DIdle(pScrn);
+}
+
 /* Set up for transparency
  *
  * Mmmm, Seems as though the transparency compare is opposite to r128.
@@ -113,14 +125,14 @@ struct R5xxXaaPrivate {
  * of that.
  */
 static void
-R5xxXAASetTransparency(ScrnInfoPtr pScrn, int trans_color)
+R5xxXAASetTransparency(struct RhdCS *CS, int trans_color)
 {
-    R5xxFIFOWait(pScrn->scrnIndex, 3);
+    RHDCSGrab(CS, 2 * 3);
 
-    RHDRegWrite(pScrn, R5XX_CLR_CMP_CLR_SRC, trans_color);
-    RHDRegWrite(pScrn, R5XX_CLR_CMP_MASK, R5XX_CLR_CMP_MSK);
-    RHDRegWrite(pScrn, R5XX_CLR_CMP_CNTL,
-		R5XX_SRC_CMP_EQ_COLOR | R5XX_CLR_CMP_SRC_SOURCE);
+    RHDCSRegWrite(CS, R5XX_CLR_CMP_CLR_SRC, trans_color);
+    RHDCSRegWrite(CS, R5XX_CLR_CMP_MASK, R5XX_CLR_CMP_MSK);
+    RHDCSRegWrite(CS, R5XX_CLR_CMP_CNTL,
+		  R5XX_SRC_CMP_EQ_COLOR | R5XX_CLR_CMP_SRC_SOURCE);
 }
 
 /*
@@ -130,6 +142,7 @@ static void
 R5xxXAASetClippingRectangle(ScrnInfoPtr pScrn, int xa, int ya, int xb, int yb)
 {
     struct R5xxXaaPrivate *XaaPrivate = RHDPTR(pScrn)->TwoDPrivate;
+    struct RhdCS *CS = RHDPTR(pScrn)->CS;
     unsigned long tmp1, tmp2;
 
     if (xa < 0) {
@@ -139,10 +152,10 @@ R5xxXAASetClippingRectangle(ScrnInfoPtr pScrn, int xa, int ya, int xb, int yb)
 	tmp1 = xa;
 
     if (ya < 0) {
-	tmp1 |= (((-ya) & 0x3fff) << 16);
+	tmp1 |= ((-ya) & 0x3fff) << 16;
 	tmp1 |= R5XX_SC_SIGN_MASK_HI;
     } else
-	tmp1 |= (ya << 16);
+	tmp1 |= ya << 16;
 
     xb++;
     yb++;
@@ -154,20 +167,22 @@ R5xxXAASetClippingRectangle(ScrnInfoPtr pScrn, int xa, int ya, int xb, int yb)
 	tmp2 = xb;
 
     if (yb < 0) {
-	tmp2 |= (((-yb) & 0x3fff) << 16);
+	tmp2 |= ((-yb) & 0x3fff) << 16;
 	tmp2 |= R5XX_SC_SIGN_MASK_HI;
     } else
-	tmp2 |= (yb << 16);
+	tmp2 |= yb << 16;
 
-    R5xxFIFOWait(pScrn->scrnIndex, 3);
+    RHDCSGrab(CS, 2 * 3);
 
-    RHDRegWrite(pScrn, R5XX_DP_GUI_MASTER_CNTL, XaaPrivate->control_saved |
-		  R5XX_GMC_DST_CLIPPING);
-    RHDRegWrite(pScrn, R5XX_SC_TOP_LEFT, tmp1);
-    RHDRegWrite(pScrn, R5XX_SC_BOTTOM_RIGHT, tmp2);
+    RHDCSRegWrite(CS, R5XX_DP_GUI_MASTER_CNTL,
+		  XaaPrivate->control_saved | R5XX_GMC_DST_CLIPPING);
+    RHDCSRegWrite(CS, R5XX_SC_TOP_LEFT, tmp1);
+    RHDCSRegWrite(CS, R5XX_SC_BOTTOM_RIGHT, tmp2);
 
     if (XaaPrivate->trans_color != -1)
-	R5xxXAASetTransparency(pScrn, XaaPrivate->trans_color);
+	R5xxXAASetTransparency(CS, XaaPrivate->trans_color);
+
+    RHDCSAdvance(CS);
 }
 
 /*
@@ -177,15 +192,18 @@ static void
 R5xxXAADisableClipping(ScrnInfoPtr pScrn)
 {
     struct R5xxXaaPrivate *XaaPrivate = RHDPTR(pScrn)->TwoDPrivate;
+    struct RhdCS *CS = RHDPTR(pScrn)->CS;
 
-    R5xxFIFOWait(pScrn->scrnIndex, 3);
+    RHDCSGrab(CS, 2 * 3);
 
-    RHDRegWrite(pScrn, R5XX_DP_GUI_MASTER_CNTL, XaaPrivate->control_saved);
-    RHDRegWrite(pScrn, R5XX_SC_TOP_LEFT, 0);
-    RHDRegWrite(pScrn, R5XX_SC_BOTTOM_RIGHT,
+    RHDCSRegWrite(CS, R5XX_DP_GUI_MASTER_CNTL, XaaPrivate->control_saved);
+    RHDCSRegWrite(CS, R5XX_SC_TOP_LEFT, 0);
+    RHDCSRegWrite(CS, R5XX_SC_BOTTOM_RIGHT,
 		  R5XX_DEFAULT_SC_RIGHT_MAX | R5XX_DEFAULT_SC_BOTTOM_MAX);
     if (XaaPrivate->trans_color != -1)
-	R5xxXAASetTransparency(pScrn, XaaPrivate->trans_color);
+	R5xxXAASetTransparency(CS, XaaPrivate->trans_color);
+
+    RHDCSAdvance(CS);
 }
 
 /*
@@ -196,6 +214,7 @@ R5xxXAASetupForSolidFill(ScrnInfoPtr pScrn,
 			 int color, int rop, unsigned int planemask)
 {
     struct R5xxXaaPrivate *XaaPrivate = RHDPTR(pScrn)->TwoDPrivate;
+    struct RhdCS *CS = RHDPTR(pScrn)->CS;
     CARD32 control;
 
     control = XaaPrivate->control | R5xxRops[rop].pattern;
@@ -204,13 +223,13 @@ R5xxXAASetupForSolidFill(ScrnInfoPtr pScrn,
     /* Save for later clipping */
     XaaPrivate->control_saved = control;
 
-    R5xxFIFOWait(pScrn->scrnIndex, 4);
+    RHDCSGrab(CS, 2 * 4);
 
-    RHDRegWrite(pScrn, R5XX_DP_GUI_MASTER_CNTL, control);
-    RHDRegWrite(pScrn, R5XX_DP_BRUSH_FRGD_CLR, color);
-    RHDRegWrite(pScrn, R5XX_DP_WRITE_MASK, planemask);
-    RHDRegWrite(pScrn, R5XX_DP_CNTL,
-		R5XX_DST_X_LEFT_TO_RIGHT | R5XX_DST_Y_TOP_TO_BOTTOM);
+    RHDCSRegWrite(CS, R5XX_DP_GUI_MASTER_CNTL, control);
+    RHDCSRegWrite(CS, R5XX_DP_BRUSH_FRGD_CLR, color);
+    RHDCSRegWrite(CS, R5XX_DP_WRITE_MASK, planemask);
+    RHDCSRegWrite(CS, R5XX_DP_CNTL,
+		  R5XX_DST_X_LEFT_TO_RIGHT | R5XX_DST_Y_TOP_TO_BOTTOM);
 }
 
 /*
@@ -220,46 +239,51 @@ static void
 R5xxXAASubsequentSolidFillRect(ScrnInfoPtr pScrn, int x, int y, int w, int h)
 {
     struct R5xxXaaPrivate *XaaPrivate = RHDPTR(pScrn)->TwoDPrivate;
+    struct RhdCS *CS = RHDPTR(pScrn)->CS;
 
-    R5xxFIFOWait(pScrn->scrnIndex, 3);
+    RHDCSGrab(CS, 2 * 3);
 
-    RHDRegWrite(pScrn, R5XX_DST_PITCH_OFFSET, XaaPrivate->dst_pitch_offset);
-    RHDRegWrite(pScrn, R5XX_DST_Y_X, (y << 16) | x);
-    RHDRegWrite(pScrn, R5XX_DST_WIDTH_HEIGHT, (w << 16) | h);
+    RHDCSRegWrite(CS, R5XX_DST_PITCH_OFFSET, XaaPrivate->dst_pitch_offset);
+    RHDCSRegWrite(CS, R5XX_DST_Y_X, (y << 16) | x);
+    RHDCSRegWrite(CS, R5XX_DST_WIDTH_HEIGHT, (w << 16) | h);
+
+    RHDCSAdvance(CS);
 }
 
 /*
- * Setup for XAA solid lines
+ *
  */
 static void
 R5xxXAASetupForSolidLine(ScrnInfoPtr pScrn,
 			 int color, int rop, unsigned int planemask)
 {
     struct R5xxXaaPrivate *XaaPrivate = RHDPTR(pScrn)->TwoDPrivate;
-    CARD32 control = XaaPrivate->control;
+    struct RhdCS *CS = RHDPTR(pScrn)->CS;
+    CARD32 control;
 
-    control |= R5XX_GMC_BRUSH_SOLID_COLOR |
-	R5XX_GMC_SRC_DATATYPE_COLOR | R5xxRops[rop].pattern;
+    control = XaaPrivate->control | R5xxRops[rop].pattern;
+    control |= R5XX_GMC_BRUSH_SOLID_COLOR | R5XX_GMC_SRC_DATATYPE_COLOR;
 
     /* Save for later clipping */
     XaaPrivate->control_saved = control;
 
-    R5xxFIFOWait(pScrn->scrnIndex, 4);
+    RHDCSGrab(CS, 2 * 4);
 
-    RHDRegWrite(pScrn, R5XX_DST_LINE_PATCOUNT, 0x55 << R5XX_BRES_CNTL_SHIFT);
-    RHDRegWrite(pScrn, R5XX_DP_GUI_MASTER_CNTL, control);
-    RHDRegWrite(pScrn, R5XX_DP_BRUSH_FRGD_CLR, color);
-    RHDRegWrite(pScrn, R5XX_DP_WRITE_MASK, planemask);
+    RHDCSRegWrite(CS, R5XX_DST_LINE_PATCOUNT, 0x55 << R5XX_BRES_CNTL_SHIFT);
+    RHDCSRegWrite(CS, R5XX_DP_GUI_MASTER_CNTL, control);
+    RHDCSRegWrite(CS, R5XX_DP_BRUSH_FRGD_CLR, color);
+    RHDCSRegWrite(CS, R5XX_DP_WRITE_MASK, planemask);
 }
 
 /*
- * Subsequent XAA solid horizontal and vertical lines
+ *
  */
 static void
 R5xxXAASubsequentSolidHorVertLine(ScrnInfoPtr pScrn,
 				  int x, int y, int len, int dir)
 {
     struct R5xxXaaPrivate *XaaPrivate = RHDPTR(pScrn)->TwoDPrivate;
+    struct RhdCS *CS = RHDPTR(pScrn)->CS;
     int w = 1, h = 1;
 
     if (dir == DEGREES_0)
@@ -267,17 +291,18 @@ R5xxXAASubsequentSolidHorVertLine(ScrnInfoPtr pScrn,
     else
 	h = len;
 
-    R5xxFIFOWait(pScrn->scrnIndex, 4);
+    RHDCSGrab(CS, 2 * 4);
 
-    RHDRegWrite(pScrn, R5XX_DP_CNTL,
-		R5XX_DST_X_LEFT_TO_RIGHT | R5XX_DST_Y_TOP_TO_BOTTOM);
-    RHDRegWrite(pScrn, R5XX_DST_PITCH_OFFSET, XaaPrivate->dst_pitch_offset);
-    RHDRegWrite(pScrn, R5XX_DST_Y_X, (y << 16) | x);
-    RHDRegWrite(pScrn, R5XX_DST_WIDTH_HEIGHT, (w << 16) | h);
+    RHDCSRegWrite(CS, R5XX_DP_CNTL,
+		  R5XX_DST_X_LEFT_TO_RIGHT | R5XX_DST_Y_TOP_TO_BOTTOM);
+    RHDCSRegWrite(CS, R5XX_DST_PITCH_OFFSET, XaaPrivate->dst_pitch_offset);
+    RHDCSRegWrite(CS, R5XX_DST_Y_X, (y << 16) | x);
+    RHDCSRegWrite(CS, R5XX_DST_WIDTH_HEIGHT, (w << 16) | h);
+
+    RHDCSAdvance(CS);
 }
 
-/* Subsequent XAA solid TwoPointLine line
- *
+/*
  * Tests: xtest CH06/drwln, ico, Mark Vojkovich's linetest program
  *
  * [See http://www.xfree86.org/devel/archives/devel/1999-Jun/0102.shtml for
@@ -288,15 +313,18 @@ R5xxXAASubsequentSolidTwoPointLine(ScrnInfoPtr pScrn, int xa, int ya,
 				   int xb, int yb, int flags)
 {
     struct R5xxXaaPrivate *XaaPrivate = RHDPTR(pScrn)->TwoDPrivate;
+    struct RhdCS *CS = RHDPTR(pScrn)->CS;
 
     if (!(flags & OMIT_LAST))
 	R5xxXAASubsequentSolidHorVertLine(pScrn, xb, yb, 1, DEGREES_0);
 
-    R5xxFIFOWait(pScrn->scrnIndex, 3);
+    RHDCSGrab(CS, 2 * 3);
 
-    RHDRegWrite(pScrn, R5XX_DST_PITCH_OFFSET, XaaPrivate->dst_pitch_offset);
-    RHDRegWrite(pScrn, R5XX_DST_LINE_START, (ya << 16) | xa);
-    RHDRegWrite(pScrn, R5XX_DST_LINE_END, (yb << 16) | xb);
+    RHDCSRegWrite(CS, R5XX_DST_PITCH_OFFSET, XaaPrivate->dst_pitch_offset);
+    RHDCSRegWrite(CS, R5XX_DST_LINE_START, (ya << 16) | xa);
+    RHDCSRegWrite(CS, R5XX_DST_LINE_END, (yb << 16) | xb);
+
+    RHDCSAdvance(CS);
 }
 
 /*
@@ -307,6 +335,7 @@ R5xxXAASetupForScreenToScreenCopy(ScrnInfoPtr pScrn, int xdir, int ydir, int rop
 				  unsigned int planemask, int trans_color)
 {
     struct R5xxXaaPrivate *XaaPrivate = RHDPTR(pScrn)->TwoDPrivate;
+    struct RhdCS *CS = RHDPTR(pScrn)->CS;
     CARD32 control;
 
     XaaPrivate->xdir = xdir;
@@ -320,17 +349,17 @@ R5xxXAASetupForScreenToScreenCopy(ScrnInfoPtr pScrn, int xdir, int ydir, int rop
     /* Save for later clipping */
     XaaPrivate->control_saved = control;
 
-    R5xxFIFOWait(pScrn->scrnIndex, 3);
+    RHDCSGrab(CS, 2 * 3);
 
-    RHDRegWrite(pScrn, R5XX_DP_GUI_MASTER_CNTL, control);
-    RHDRegWrite(pScrn, R5XX_DP_WRITE_MASK, planemask);
-    RHDRegWrite(pScrn, R5XX_DP_CNTL,
-		((xdir >= 0 ? R5XX_DST_X_LEFT_TO_RIGHT : 0) |
-		 (ydir >= 0 ? R5XX_DST_Y_TOP_TO_BOTTOM : 0)));
+    RHDCSRegWrite(CS, R5XX_DP_GUI_MASTER_CNTL, control);
+    RHDCSRegWrite(CS, R5XX_DP_WRITE_MASK, planemask);
+    RHDCSRegWrite(CS, R5XX_DP_CNTL,
+		  (xdir >= 0 ? R5XX_DST_X_LEFT_TO_RIGHT : 0) |
+		  (ydir >= 0 ? R5XX_DST_Y_TOP_TO_BOTTOM : 0));
 
     XaaPrivate->trans_color = trans_color;
     if (trans_color != -1)
-	R5xxXAASetTransparency(pScrn, trans_color);
+	R5xxXAASetTransparency(CS, trans_color);
 }
 
 /*
@@ -341,6 +370,7 @@ R5xxXAASubsequentScreenToScreenCopy(ScrnInfoPtr pScrn, int xa, int ya,
 				    int xb, int yb, int w, int h)
 {
     struct R5xxXaaPrivate *XaaPrivate = RHDPTR(pScrn)->TwoDPrivate;
+    struct RhdCS *CS = RHDPTR(pScrn)->CS;
 
     if (XaaPrivate->xdir < 0) {
 	xa += w - 1;
@@ -352,16 +382,19 @@ R5xxXAASubsequentScreenToScreenCopy(ScrnInfoPtr pScrn, int xa, int ya,
 	yb += h - 1;
     }
 
-    R5xxFIFOWait(pScrn->scrnIndex, 5);
+    RHDCSGrab(CS, 2 * 5);
 
-    RHDRegWrite(pScrn, R5XX_SRC_PITCH_OFFSET, XaaPrivate->dst_pitch_offset);
-    RHDRegWrite(pScrn, R5XX_DST_PITCH_OFFSET, XaaPrivate->dst_pitch_offset);
-    RHDRegWrite(pScrn, R5XX_SRC_Y_X, (ya << 16) | xa);
-    RHDRegWrite(pScrn, R5XX_DST_Y_X, (yb << 16) | xb);
-    RHDRegWrite(pScrn, R5XX_DST_HEIGHT_WIDTH, (h << 16) | w);
+    RHDCSRegWrite(CS, R5XX_SRC_PITCH_OFFSET, XaaPrivate->dst_pitch_offset);
+    RHDCSRegWrite(CS, R5XX_DST_PITCH_OFFSET, XaaPrivate->dst_pitch_offset);
+    RHDCSRegWrite(CS, R5XX_SRC_Y_X, (ya << 16) | xa);
+    RHDCSRegWrite(CS, R5XX_DST_Y_X, (yb << 16) | xb);
+    RHDCSRegWrite(CS, R5XX_DST_HEIGHT_WIDTH, (h << 16) | w);
+
+    RHDCSAdvance(CS);
 }
 
-/* Setup for XAA mono 8x8 pattern color expansion.  Patterns with
+/*
+ * Setup for XAA mono 8x8 pattern color expansion.  Patterns with
  * transparency use `bg == -1'.  This routine is only used if the XAA
  * pixmap cache is turned on.
  *
@@ -373,6 +406,7 @@ R5xxXAASetupForMono8x8PatternFill(ScrnInfoPtr pScrn, int patternx, int patterny,
 				  int fg, int bg, int rop, unsigned int planemask)
 {
     struct R5xxXaaPrivate *XaaPrivate = RHDPTR(pScrn)->TwoDPrivate;
+    struct RhdCS *CS = RHDPTR(pScrn)->CS;
     CARD32 control;
 
 #if X_BYTE_ORDER == X_BIG_ENDIAN
@@ -397,21 +431,22 @@ R5xxXAASetupForMono8x8PatternFill(ScrnInfoPtr pScrn, int patternx, int patterny,
     XaaPrivate->control_saved = control;
 
     if (bg != -1)
-	R5xxFIFOWait(pScrn->scrnIndex, 6);
+	RHDCSGrab(CS, 2 * 6);
     else
-	R5xxFIFOWait(pScrn->scrnIndex, 5);
+	RHDCSGrab(CS, 2 * 5);
 
-    RHDRegWrite(pScrn, R5XX_DP_GUI_MASTER_CNTL, control);
-    RHDRegWrite(pScrn, R5XX_DP_WRITE_MASK, planemask);
-    RHDRegWrite(pScrn, R5XX_DP_BRUSH_FRGD_CLR, fg);
+    RHDCSRegWrite(CS, R5XX_DP_GUI_MASTER_CNTL, control);
+    RHDCSRegWrite(CS, R5XX_DP_WRITE_MASK, planemask);
+    RHDCSRegWrite(CS, R5XX_DP_BRUSH_FRGD_CLR, fg);
     if (bg != -1)
-	RHDRegWrite(pScrn, R5XX_DP_BRUSH_BKGD_CLR, bg);
+	RHDCSRegWrite(CS, R5XX_DP_BRUSH_BKGD_CLR, bg);
 
-    RHDRegWrite(pScrn, R5XX_BRUSH_DATA0, patternx);
-    RHDRegWrite(pScrn, R5XX_BRUSH_DATA1, patterny);
+    RHDCSRegWrite(CS, R5XX_BRUSH_DATA0, patternx);
+    RHDCSRegWrite(CS, R5XX_BRUSH_DATA1, patterny);
 }
 
-/* Subsequent XAA 8x8 pattern color expansion.  Because they are used in
+/*
+ * Subsequent XAA 8x8 pattern color expansion.  Because they are used in
  * the setup function, `patternx' and `patterny' are not used here.
  */
 static void
@@ -420,26 +455,31 @@ R5xxXAASubsequentMono8x8PatternFillRect(ScrnInfoPtr pScrn,
 					int x, int y, int w, int h)
 {
     struct R5xxXaaPrivate *XaaPrivate = RHDPTR(pScrn)->TwoDPrivate;
+    struct RhdCS *CS = RHDPTR(pScrn)->CS;
 
-    R5xxFIFOWait(pScrn->scrnIndex, 4);
+    RHDCSGrab(CS, 2 * 4);
 
-    RHDRegWrite(pScrn, R5XX_DST_PITCH_OFFSET, XaaPrivate->dst_pitch_offset);
-    RHDRegWrite(pScrn, R5XX_BRUSH_Y_X, (patterny << 8) | patternx);
-    RHDRegWrite(pScrn, R5XX_DST_Y_X, (y << 16) | x);
-    RHDRegWrite(pScrn, R5XX_DST_HEIGHT_WIDTH, (h << 16) | w);
+    RHDCSRegWrite(CS, R5XX_DST_PITCH_OFFSET, XaaPrivate->dst_pitch_offset);
+    RHDCSRegWrite(CS, R5XX_BRUSH_Y_X, (patterny << 8) | patternx);
+    RHDCSRegWrite(CS, R5XX_DST_Y_X, (y << 16) | x);
+    RHDCSRegWrite(CS, R5XX_DST_HEIGHT_WIDTH, (h << 16) | w);
+
+    RHDCSAdvance(CS);
 }
 
-/* Setup for XAA indirect CPU-to-screen color expansion (indirect).
+/*
+ * Setup for XAA indirect CPU-to-screen color expansion (indirect).
  * Because of how the scratch buffer is initialized, this is really a
  * mainstore-to-screen color expansion.  Transparency is supported when
  * `bg == -1'.
  * We always need to provide the bg here, otherwise the engine locks.
  */
 static void
-R5xxXAASetupForScanlineCPUToScreenColorExpandFill(ScrnInfoPtr pScrn, int fg, int bg,
-						  int rop, unsigned int planemask)
+R5xxXAASetupForScanlineCPUToScreenColorExpandFillMMIO(ScrnInfoPtr pScrn, int fg, int bg,
+						      int rop, unsigned int planemask)
 {
     struct R5xxXaaPrivate *XaaPrivate = RHDPTR(pScrn)->TwoDPrivate;
+    struct RhdCS *CS = RHDPTR(pScrn)->CS;
     CARD32 control = XaaPrivate->control;
 
     XaaPrivate->scanline_bpp = 0;
@@ -462,47 +502,53 @@ R5xxXAASetupForScanlineCPUToScreenColorExpandFill(ScrnInfoPtr pScrn, int fg, int
     XaaPrivate->control_saved = control;
 
 #if X_BYTE_ORDER == X_LITTLE_ENDIAN
-    R5xxFIFOWait(pScrn->scrnIndex, 4);
+    RHDCSGrab(CS, 2 * 4);
 #else
-    R5xxFIFOWait(pScrn->scrnIndex, 5);
+    RHDCSGrab(CS, 2 * 5);
 
-    RHDRegWrite(pScrn, R5XX_RBBM_GUICNTL, R5XX_HOST_DATA_SWAP_NONE);
+    RHDCSRegWrite(CS, R5XX_RBBM_GUICNTL, R5XX_HOST_DATA_SWAP_NONE);
 #endif
-    RHDRegWrite(pScrn, R5XX_DP_GUI_MASTER_CNTL, control);
-    RHDRegWrite(pScrn, R5XX_DP_WRITE_MASK, planemask);
-    RHDRegWrite(pScrn, R5XX_DP_SRC_FRGD_CLR, fg);
-    RHDRegWrite(pScrn, R5XX_DP_SRC_BKGD_CLR, bg);
+    RHDCSRegWrite(CS, R5XX_DP_GUI_MASTER_CNTL, control);
+    RHDCSRegWrite(CS, R5XX_DP_WRITE_MASK, planemask);
+    RHDCSRegWrite(CS, R5XX_DP_SRC_FRGD_CLR, fg);
+    RHDCSRegWrite(CS, R5XX_DP_SRC_BKGD_CLR, bg);
 }
 
-/* Subsequent XAA indirect CPU-to-screen color expansion.  This is only
+/*
+ * Subsequent XAA indirect CPU-to-screen color expansion.  This is only
  * called once for each rectangle.
  */
 static void
-R5xxXAASubsequentScanlineCPUToScreenColorExpandFill(ScrnInfoPtr pScrn, int x, int y,
-						    int w, int h, int skipleft)
+R5xxXAASubsequentScanlineCPUToScreenColorExpandFillMMIO(ScrnInfoPtr pScrn, int x, int y,
+							int w, int h, int skipleft)
 {
     struct R5xxXaaPrivate *XaaPrivate = RHDPTR(pScrn)->TwoDPrivate;
+    struct RhdCS *CS = RHDPTR(pScrn)->CS;
 
     XaaPrivate->scanline_h = h;
     XaaPrivate->scanline_words = (w + 31) >> 5;
 
-    R5xxFIFOWait(pScrn->scrnIndex, 5);
+    RHDCSGrab(CS, 2 * 5);
 
-    RHDRegWrite(pScrn, R5XX_DST_PITCH_OFFSET, XaaPrivate->dst_pitch_offset);
-    RHDRegWrite(pScrn, R5XX_SC_TOP_LEFT, (y << 16) | ((x + skipleft) & 0xffff));
-    RHDRegWrite(pScrn, R5XX_SC_BOTTOM_RIGHT, ((y + h) << 16) | ((x + w) & 0xffff));
-    RHDRegWrite(pScrn, R5XX_DST_Y_X, (y << 16) | (x & 0xffff));
+    RHDCSRegWrite(CS, R5XX_DST_PITCH_OFFSET, XaaPrivate->dst_pitch_offset);
+    RHDCSRegWrite(CS, R5XX_SC_TOP_LEFT, (y << 16) | ((x + skipleft) & 0xffff));
+    RHDCSRegWrite(CS, R5XX_SC_BOTTOM_RIGHT, ((y + h) << 16) | ((x + w) & 0xffff));
+    RHDCSRegWrite(CS, R5XX_DST_Y_X, (y << 16) | (x & 0xffff));
     /* Have to pad the width here and use clipping engine */
-    RHDRegWrite(pScrn, R5XX_DST_HEIGHT_WIDTH, (h << 16) | ((w + 31) & ~31));
+    RHDCSRegWrite(CS, R5XX_DST_HEIGHT_WIDTH, (h << 16) | ((w + 31) & ~31));
+
+    RHDCSAdvance(CS);
 }
 
-/* Subsequent XAA indirect CPU-to-screen color expansion and indirect
+/*
+ * Subsequent XAA indirect CPU-to-screen color expansion and indirect
  * image write.  This is called once for each scanline.
  */
 static void
-R5xxXAASubsequentScanline(ScrnInfoPtr pScrn, int bufno)
+R5xxXAASubsequentScanlineMMIO(ScrnInfoPtr pScrn, int bufno)
 {
     struct R5xxXaaPrivate *XaaPrivate = RHDPTR(pScrn)->TwoDPrivate;
+    struct RhdCS *CS = RHDPTR(pScrn)->CS;
     CARD32 *p = (CARD32 *) XaaPrivate->BufferHook[bufno];
     int i, remainder, left = XaaPrivate->scanline_words;
     CARD16 Reg;
@@ -517,9 +563,9 @@ R5xxXAASubsequentScanline(ScrnInfoPtr pScrn, int bufno)
     while (left > remainder) {
 	Reg = R5XX_HOST_DATA0;
 
-	R5xxFIFOWait(pScrn->scrnIndex, 8);
+	RHDCSGrab(CS, 2 * 8);
 	for (i = 0; i < 8; i++, Reg += 4)
-	    RHDRegWrite(pScrn, Reg, *p++);
+	    RHDCSRegWrite(CS, Reg, *p++);
 
 	left -= 8;
     }
@@ -531,19 +577,23 @@ R5xxXAASubsequentScanline(ScrnInfoPtr pScrn, int bufno)
 
     Reg -= 4 * (left - 1);
 
-    R5xxFIFOWait(pScrn->scrnIndex, left);
+    RHDCSGrab(CS, 2 * left);
     for (i = 0; i < left; i++, Reg += 4)
-        RHDRegWrite(pScrn, Reg, *p++);
+	RHDCSRegWrite(CS, Reg, *p++);
+
+    if (!(XaaPrivate->scanline_h % 32)) /* don't push too big an amount of data */
+	RHDCSAdvance(CS);
 }
 
 /*
  * Setup for XAA indirect image write
  */
 static void
-R5xxXAASetupForScanlineImageWrite(ScrnInfoPtr pScrn, int rop, unsigned int planemask,
-				  int trans_color, int bpp, int depth)
+R5xxXAASetupForScanlineImageWriteMMIO(ScrnInfoPtr pScrn, int rop, unsigned int planemask,
+				      int trans_color, int bpp, int depth)
 {
     struct R5xxXaaPrivate *XaaPrivate = RHDPTR(pScrn)->TwoDPrivate;
+    struct RhdCS *CS = RHDPTR(pScrn)->CS;
     CARD32 control = XaaPrivate->control;
 
     XaaPrivate->scanline_bpp = bpp;
@@ -556,33 +606,34 @@ R5xxXAASetupForScanlineImageWrite(ScrnInfoPtr pScrn, int rop, unsigned int plane
     XaaPrivate->control_saved = control;
 
 #if X_BYTE_ORDER == X_LITTLE_ENDIAN
-    R5xxFIFOWait(pScrn->scrnIndex, 2);
+    RHDCSGrab(CS, 2 * 2);
 #else
-    R5xxFIFOWait(pScrn->scrnIndex, 3);
+    RHDCSGrab(CS, 2 * 3);
 
     if (bpp == 16)
-	RHDRegWrite(pScrn, R5XX_RBBM_GUICNTL, R5XX_HOST_DATA_SWAP_16BIT);
+	RHDCSRegWrite(CS, R5XX_RBBM_GUICNTL, R5XX_HOST_DATA_SWAP_16BIT);
     else if (bpp == 32)
-	RHDRegWrite(pScrn, R5XX_RBBM_GUICNTL, R5XX_HOST_DATA_SWAP_32BIT);
+	RHDCSRegWrite(CS, R5XX_RBBM_GUICNTL, R5XX_HOST_DATA_SWAP_32BIT);
     else
-	RHDRegWrite(pScrn, R5XX_RBBM_GUICNTL, R5XX_HOST_DATA_SWAP_NONE);
+	RHDCSRegWrite(CS, R5XX_RBBM_GUICNTL, R5XX_HOST_DATA_SWAP_NONE);
 #endif
-    RHDRegWrite(pScrn, R5XX_DP_GUI_MASTER_CNTL, control);
-    RHDRegWrite(pScrn, R5XX_DP_WRITE_MASK, planemask);
+    RHDCSRegWrite(CS, R5XX_DP_GUI_MASTER_CNTL, control);
+    RHDCSRegWrite(CS, R5XX_DP_WRITE_MASK, planemask);
 
     XaaPrivate->trans_color = trans_color;
     if (trans_color != -1)
-	R5xxXAASetTransparency(pScrn, trans_color);
+	R5xxXAASetTransparency(CS, trans_color);
 }
 
 /*
  *
  */
 static void
-R5xxXAASubsequentScanlineImageWriteRect(ScrnInfoPtr pScrn, int x, int y,
-					int w, int h, int skipleft)
+R5xxXAASubsequentScanlineImageWriteRectMMIO(ScrnInfoPtr pScrn, int x, int y,
+					    int w, int h, int skipleft)
 {
     struct R5xxXaaPrivate *XaaPrivate = RHDPTR(pScrn)->TwoDPrivate;
+    struct RhdCS *CS = RHDPTR(pScrn)->CS;
     int shift = 0; /* 32bpp */
 
     if (pScrn->bitsPerPixel == 8)
@@ -593,13 +644,251 @@ R5xxXAASubsequentScanlineImageWriteRect(ScrnInfoPtr pScrn, int x, int y,
     XaaPrivate->scanline_h = h;
     XaaPrivate->scanline_words  = (w * XaaPrivate->scanline_bpp + 31) >> 5;
 
-    R5xxFIFOWait(pScrn->scrnIndex, 5);
-    RHDRegWrite(pScrn, R5XX_DST_PITCH_OFFSET, XaaPrivate->dst_pitch_offset);
-    RHDRegWrite(pScrn, R5XX_SC_TOP_LEFT, (y << 16) | ((x + skipleft) & 0xffff));
-    RHDRegWrite(pScrn, R5XX_SC_BOTTOM_RIGHT, ((y + h) << 16) | ((x + w) & 0xffff));
-    RHDRegWrite(pScrn, R5XX_DST_Y_X, (y << 16) | (x & 0xffff));
+    RHDCSGrab(CS, 2 * 5);
+    RHDCSRegWrite(CS, R5XX_DST_PITCH_OFFSET, XaaPrivate->dst_pitch_offset);
+    RHDCSRegWrite(CS, R5XX_SC_TOP_LEFT, (y << 16) | ((x + skipleft) & 0xffff));
+    RHDCSRegWrite(CS, R5XX_SC_BOTTOM_RIGHT, ((y + h) << 16) | ((x + w) & 0xffff));
+    RHDCSRegWrite(CS, R5XX_DST_Y_X, (y << 16) | (x & 0xffff));
     /* Have to pad the width here and use clipping engine */
-    RHDRegWrite(pScrn, R5XX_DST_HEIGHT_WIDTH, (h << 16) | ((w + shift) & ~shift));
+    RHDCSRegWrite(CS, R5XX_DST_HEIGHT_WIDTH, (h << 16) | ((w + shift) & ~shift));
+
+    RHDCSAdvance(CS);
+}
+
+/*
+ * Helper function to write out a HOSTDATA_BLT packet into the indirect
+ * buffer and set the XAA scratch buffer address appropriately.
+ */
+static void
+R5xxXAACPScanlinePacket(struct RhdCS *CS, struct R5xxXaaPrivate *XaaPrivate)
+{
+    int chunk_words = XaaPrivate->scanline_hpass * XaaPrivate->scanline_words;
+
+    RHDCSGrab(CS, chunk_words + 10);
+
+    RHDCSWrite(CS, CP_PACKET3(R5XX_CP_PACKET3_CNTL_HOSTDATA_BLT,chunk_words+10-2));
+    RHDCSWrite(CS, XaaPrivate->control_saved);
+    RHDCSWrite(CS, XaaPrivate->dst_pitch_offset);
+    RHDCSWrite(CS, (XaaPrivate->scanline_y << 16) |
+	       (XaaPrivate->scanline_x1clip & 0xffff));
+    RHDCSWrite(CS, ((XaaPrivate->scanline_y + XaaPrivate->scanline_hpass) << 16) |
+	       (XaaPrivate->scanline_x2clip & 0xffff));
+    RHDCSWrite(CS, XaaPrivate->scanline_fg);
+    RHDCSWrite(CS, XaaPrivate->scanline_bg);
+    RHDCSWrite(CS, (XaaPrivate->scanline_y << 16) |
+	       (XaaPrivate->scanline_x & 0xffff));
+    RHDCSWrite(CS, (XaaPrivate->scanline_hpass << 16) |
+	       (XaaPrivate->scanline_w & 0xffff));
+    RHDCSWrite(CS, chunk_words);
+
+    XaaPrivate->BufferHook[0] = (CARD8 *) &(CS->Buffer[CS->Wptr]);
+    CS->Wptr += chunk_words;
+
+    XaaPrivate->scanline_y += XaaPrivate->scanline_hpass;
+    XaaPrivate->scanline_h -= XaaPrivate->scanline_hpass;
+}
+
+/*
+ * Setup for XAA indirect CPU-to-screen color expansion (indirect).
+ * Because of how the scratch buffer is initialized, this is really a
+ * mainstore-to-screen color expansion.  Transparency is supported when
+ * `bg == -1'.
+ */
+static void
+R5xxXAASetupForScanlineCPUToScreenColorExpandFillCP(ScrnInfoPtr pScrn, int fg,
+						    int bg, int rop, unsigned int planemask)
+{
+    struct R5xxXaaPrivate *XaaPrivate = RHDPTR(pScrn)->TwoDPrivate;
+    struct RhdCS *CS = RHDPTR(pScrn)->CS;
+    CARD32 control = XaaPrivate->control;
+
+    XaaPrivate->scanline_bpp = 0;
+
+    control |= R5XX_GMC_DST_CLIPPING | R5XX_GMC_BRUSH_NONE |
+	R5xxRops[rop].rop | R5XX_DP_SRC_SOURCE_HOST_DATA;
+
+    if (bg == -1)
+	control |= R5XX_GMC_SRC_DATATYPE_MONO_FG_LA;
+    else
+	control |= R5XX_GMC_SRC_DATATYPE_MONO_FG_BG;
+
+#if X_BYTE_ORDER == X_LITTLE_ENDIAN
+    control |= R5XX_GMC_BYTE_LSB_TO_MSB;
+#else
+    control |= R5XX_GMC_BYTE_MSB_TO_LSB;
+#endif
+
+    XaaPrivate->control_saved = control;
+
+    XaaPrivate->scanline_fg = fg;
+    XaaPrivate->scanline_bg = bg;
+
+    RHDCSGrab(CS, 2 * 1);
+    RHDCSRegWrite(CS, R5XX_DP_WRITE_MASK, planemask);
+}
+
+/*
+ * Subsequent XAA indirect CPU-to-screen color expansion.  This is only
+ * called once for each rectangle.
+ */
+static void
+R5xxXAASubsequentScanlineCPUToScreenColorExpandFillCP(ScrnInfoPtr pScrn, int x, int y,
+						      int w, int h, int skipleft)
+{
+    struct R5xxXaaPrivate *XaaPrivate = RHDPTR(pScrn)->TwoDPrivate;
+    struct RhdCS *CS = RHDPTR(pScrn)->CS;
+
+    XaaPrivate->scanline_x = x;
+    XaaPrivate->scanline_y = y;
+    /* Have to pad the width here and use clipping engine */
+    XaaPrivate->scanline_w = (w + 31) & ~31;
+    XaaPrivate->scanline_h = h;
+
+    XaaPrivate->scanline_x1clip = x + skipleft;
+    XaaPrivate->scanline_x2clip = x + w;
+
+    XaaPrivate->scanline_words = XaaPrivate->scanline_w / 32;
+    XaaPrivate->scanline_hpass =
+	min(XaaPrivate->scanline_h, ((CS->Size - 10) / XaaPrivate->scanline_words));
+
+    R5xxXAACPScanlinePacket(CS, XaaPrivate);
+}
+
+#if X_BYTE_ORDER == X_BIG_ENDIAN
+
+static inline void
+R5xxXAAScanlineSwap16(unsigned int *Buf, unsigned int nwords)
+{
+    unsigned int *d = Buf;
+    unsigned int *s = Buf;
+
+    for (; nwords > 0; --nwords, ++d, ++s)
+	*d = ((*s & 0xffff) << 16) | ((*s >> 16) & 0xffff);
+}
+
+static inline void
+R5xxXAAScanlineSwap8(unsigned int *Buf, unsigned int nwords)
+{
+    unsigned int *d = Buf;
+    unsigned int *s = Buf;
+
+    for (; nwords > 0; --nwords, ++d, ++s)
+#ifdef __powerpc__
+	asm volatile("stwbrx %0,0,%1" : : "r" (*s), "r" (d));
+#else
+    *d = ((*s >> 24) & 0xff) | ((*s >> 8) & 0xff00)
+	| ((*s & 0xff00) << 8) | ((*s & 0xff) << 24);
+#endif
+}
+
+#endif /* X_BYTE_ORDER */
+
+/*
+ * Subsequent XAA indirect CPU-to-screen color expansion and indirect
+ * image write.  This is called once for each scanline.
+ */
+static void
+R5xxXAASubsequentScanlineCP(ScrnInfoPtr pScrn, int bufno)
+{
+    struct R5xxXaaPrivate *XaaPrivate = RHDPTR(pScrn)->TwoDPrivate;
+    struct RhdCS *CS = RHDPTR(pScrn)->CS;
+
+    if (bufno) {
+	xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
+		   "%s: buffer number provided (%d)\n", __func__, bufno);
+	return;
+    }
+
+#if X_BYTE_ORDER == X_BIG_ENDIAN
+    if (XaaPrivate->scanline_bpp == 16)
+	R5xxXAAScanlineSwap16((unsigned int *)XaaPrivate->BufferHook[0],
+			     XaaPrivate->scanline_words);
+    else if (XaaPrivate->scanline_bpp < 15)
+	R5xxXAAScanlineSwap8((unsigned int *)XaaPrivate->BufferHook[0],
+			    XaaPrivate->scanline_words);
+#endif
+
+    if (--XaaPrivate->scanline_hpass) {
+	XaaPrivate->BufferHook[0] += 4 * XaaPrivate->scanline_words;
+    } else if (XaaPrivate->scanline_h) {
+	XaaPrivate->scanline_hpass =
+	    min(XaaPrivate->scanline_h, ((CS->Size - 10) / XaaPrivate->scanline_words));
+	R5xxXAACPScanlinePacket(CS, XaaPrivate);
+    }
+}
+
+/*
+ * Setup for XAA indirect image write
+ */
+static void
+R5xxXAASetupForScanlineImageWriteCP(ScrnInfoPtr pScrn, int rop, unsigned int planemask,
+				    int trans_color, int bpp, int depth)
+{
+    struct R5xxXaaPrivate *XaaPrivate = RHDPTR(pScrn)->TwoDPrivate;
+    struct RhdCS *CS = RHDPTR(pScrn)->CS;
+    CARD32 control = XaaPrivate->control;
+
+    XaaPrivate->scanline_bpp = bpp;
+
+    control |= R5XX_GMC_DST_CLIPPING | R5XX_GMC_BRUSH_NONE |
+	R5XX_GMC_SRC_DATATYPE_COLOR | R5xxRops[rop].rop |
+	R5XX_GMC_BYTE_MSB_TO_LSB | R5XX_DP_SRC_SOURCE_HOST_DATA;
+
+     /* Save for later clipping */
+    XaaPrivate->control_saved = control;
+
+    RHDCSGrab(CS, 2 * 2);
+
+    RHDCSRegWrite(CS, R5XX_DP_GUI_MASTER_CNTL, control);
+    RHDCSRegWrite(CS, R5XX_DP_WRITE_MASK, planemask);
+
+    XaaPrivate->trans_color = trans_color;
+    if (trans_color != -1)
+	R5xxXAASetTransparency(CS, trans_color);
+}
+
+/*
+ * Subsequent XAA indirect image write. This is only called once for
+ * each rectangle.
+ */
+static void
+R5xxXAASubsequentScanlineImageWriteRectCP(ScrnInfoPtr pScrn, int x, int y,
+					  int w, int h, int skipleft)
+{
+    struct R5xxXaaPrivate *XaaPrivate = RHDPTR(pScrn)->TwoDPrivate;
+    struct RhdCS *CS = RHDPTR(pScrn)->CS;
+    int pad;
+
+    switch (pScrn->bitsPerPixel) {
+    case 8:
+	pad = 3;
+	break;
+    case 16:
+	pad = 1;
+	break;
+    case 32:
+	pad = 0;
+	break;
+    default:
+	xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "%s: unhandled bpp: %d\n",
+		   __func__, pScrn->bitsPerPixel);
+	pad = 0;
+    }
+
+    XaaPrivate->scanline_x = x;
+    XaaPrivate->scanline_y = y;
+    /* Have to pad the width here and use clipping engine */
+    XaaPrivate->scanline_w = (w + pad) & ~pad;
+    XaaPrivate->scanline_h = h;
+
+    XaaPrivate->scanline_x1clip = x + skipleft;
+    XaaPrivate->scanline_x2clip = x + w;
+
+    XaaPrivate->scanline_words = (w * XaaPrivate->scanline_bpp + 31) / 32;
+    XaaPrivate->scanline_hpass =
+	min(XaaPrivate->scanline_h, ((CS->Size - 10) / XaaPrivate->scanline_words));
+
+    R5xxXAACPScanlinePacket(CS, XaaPrivate);
 }
 
 /*
@@ -609,13 +898,14 @@ static void
 R5xxXAAFunctionsInit(ScrnInfoPtr pScrn, ScreenPtr pScreen, XAAInfoRecPtr XAAInfo)
 {
     struct R5xxXaaPrivate *XaaPrivate = RHDPTR(pScrn)->TwoDPrivate;
+    struct RhdCS *CS = RHDPTR(pScrn)->CS;
 
     RHDFUNC(pScrn);
 
     XAAInfo->Flags = PIXMAP_CACHE | OFFSCREEN_PIXMAPS | LINEAR_FRAMEBUFFER;
 
     /* Sync */
-    XAAInfo->Sync = R5xx2DIdle;
+    XAAInfo->Sync = R5xxXAAIdle;
 
     /* Clipping */
     XAAInfo->SetClippingRectangle = R5xxXAASetClippingRectangle;
@@ -667,10 +957,13 @@ R5xxXAAFunctionsInit(ScrnInfoPtr pScrn, ScreenPtr pScreen, XAAInfoRecPtr XAAInfo
     XAAInfo->SetupForMono8x8PatternFill = R5xxXAASetupForMono8x8PatternFill;
     XAAInfo->SubsequentMono8x8PatternFillRect = R5xxXAASubsequentMono8x8PatternFillRect;
 
-    if (!XaaPrivate->Buffer)
-	XaaPrivate->Buffer = xnfcalloc(1, ((pScrn->virtualX + 31) / 32 * 4) +
-				       (pScrn->virtualX * (pScrn->bitsPerPixel / 8)));
-    XaaPrivate->BufferHook[0] = XaaPrivate->Buffer;
+    /* In the MMIO Case, we buffer the scanline */
+    if (CS->Type == RHD_CS_MMIO) {
+	if (!XaaPrivate->Buffer)
+	    XaaPrivate->Buffer = xnfcalloc(1, ((pScrn->virtualX + 31) / 32 * 4) +
+					   (pScrn->virtualX * (pScrn->bitsPerPixel / 8)));
+	XaaPrivate->BufferHook[0] = XaaPrivate->Buffer;
+    }
 
     /* Indirect CPU-To-Screen Color Expand
      *
@@ -681,24 +974,45 @@ R5xxXAAFunctionsInit(ScrnInfoPtr pScrn, ScreenPtr pScreen, XAAInfoRecPtr XAAInfo
 	LEFT_EDGE_CLIPPING | ROP_NEEDS_SOURCE | LEFT_EDGE_CLIPPING_NEGATIVE_X;
     XAAInfo->NumScanlineColorExpandBuffers = 1;
     XAAInfo->ScanlineColorExpandBuffers = XaaPrivate->BufferHook;
-    XAAInfo->SetupForScanlineCPUToScreenColorExpandFill = R5xxXAASetupForScanlineCPUToScreenColorExpandFill;
-    XAAInfo->SubsequentScanlineCPUToScreenColorExpandFill = R5xxXAASubsequentScanlineCPUToScreenColorExpandFill;
-    XAAInfo->SubsequentColorExpandScanline = R5xxXAASubsequentScanline;
+
+    if (CS->Type == RHD_CS_MMIO) {
+	XAAInfo->SetupForScanlineCPUToScreenColorExpandFill
+	    = R5xxXAASetupForScanlineCPUToScreenColorExpandFillMMIO;
+	XAAInfo->SubsequentScanlineCPUToScreenColorExpandFill
+	    = R5xxXAASubsequentScanlineCPUToScreenColorExpandFillMMIO;
+	XAAInfo->SubsequentColorExpandScanline = R5xxXAASubsequentScanlineMMIO;
+    } else {
+	XAAInfo->SetupForScanlineCPUToScreenColorExpandFill
+	    = R5xxXAASetupForScanlineCPUToScreenColorExpandFillCP;
+	XAAInfo->SubsequentScanlineCPUToScreenColorExpandFill
+	    = R5xxXAASubsequentScanlineCPUToScreenColorExpandFillCP;
+	XAAInfo->SubsequentColorExpandScanline = R5xxXAASubsequentScanlineCP;
+    }
 
     /* ImageWrite */
     XAAInfo->ScanlineImageWriteFlags = CPU_TRANSFER_PAD_DWORD |
-	/* Performance tests show that we shouldn't use MMIOed GXcopy for uploads
-	   as a memcpy is faster */
-	NO_GXCOPY |
 	/* R5XX gets upset, when using HOST provided data without a source rop.
 	   To show run 'xtest's ptimg */
 	ROP_NEEDS_SOURCE |
 	SCANLINE_PAD_DWORD | LEFT_EDGE_CLIPPING | LEFT_EDGE_CLIPPING_NEGATIVE_X;
+
+    /* Performance tests show that we shouldn't use MMIOed GXcopy for uploads
+       as a memcpy is faster */
+    if (CS->Type == RHD_CS_MMIO)
+	XAAInfo->ScanlineImageWriteFlags |= NO_GXCOPY;
+
     XAAInfo->NumScanlineImageWriteBuffers = 1;
     XAAInfo->ScanlineImageWriteBuffers = XaaPrivate->BufferHook;
-    XAAInfo->SetupForScanlineImageWrite = R5xxXAASetupForScanlineImageWrite;
-    XAAInfo->SubsequentScanlineImageWriteRect = R5xxXAASubsequentScanlineImageWriteRect;
-    XAAInfo->SubsequentImageWriteScanline = R5xxXAASubsequentScanline;
+
+    if (CS->Type == RHD_CS_MMIO) {
+	XAAInfo->SetupForScanlineImageWrite = R5xxXAASetupForScanlineImageWriteMMIO;
+	XAAInfo->SubsequentScanlineImageWriteRect = R5xxXAASubsequentScanlineImageWriteRectMMIO;
+	XAAInfo->SubsequentImageWriteScanline = R5xxXAASubsequentScanlineMMIO;
+    } else {
+	XAAInfo->SetupForScanlineImageWrite = R5xxXAASetupForScanlineImageWriteCP;
+	XAAInfo->SubsequentScanlineImageWriteRect = R5xxXAASubsequentScanlineImageWriteRectCP;
+	XAAInfo->SubsequentImageWriteScanline = R5xxXAASubsequentScanlineCP;
+    }
 }
 
 /*
@@ -743,11 +1057,11 @@ static void
 R5xxXaaPrivateInit(ScrnInfoPtr pScrn)
 {
     RHDPtr rhdPtr = RHDPTR(pScrn);
-    struct R5xxXaaPrivate *XaaPrivate = xnfcalloc(1, sizeof(struct R5xxXaaPrivate));
+    struct R5xxXaaPrivate *XaaPrivate =
+	xnfcalloc(1, sizeof(struct R5xxXaaPrivate));
 
-    XaaPrivate->scrnIndex = pScrn->scrnIndex;
-
-    XaaPrivate->control = (R5xx2DDatatypeGet(pScrn) << R5XX_GMC_DST_DATATYPE_SHIFT) |
+    XaaPrivate->control =
+	(R5xx2DDatatypeGet(pScrn) << R5XX_GMC_DST_DATATYPE_SHIFT) |
 	R5XX_GMC_CLR_CMP_CNTL_DIS | R5XX_GMC_DST_PITCH_OFFSET_CNTL;
 
     XaaPrivate->dst_pitch_offset =
@@ -809,8 +1123,6 @@ R5xxXAAInit(ScrnInfoPtr pScrn, ScreenPtr pScreen)
 
     rhdPtr->XAAInfo = XAAInfo;
 
-    R5xx2DStart(pScrn);
-
     return TRUE;
 }
 
