@@ -39,7 +39,8 @@ struct rhdConnectorInfo {
 enum rhdCardFlag {
     RHD_CARD_FLAG_NONE  = 0,
     RHD_CARD_FLAG_DMS59 = 1, /* DMS59 connector is only reported as two DVI-I */
-    RHD_CARD_FLAG_HPDSWAP = 2 /* some cards have broken connector tables */
+    RHD_CARD_FLAG_HPDSWAP = 2, /* some cards have broken connector tables */
+    RHD_CARD_FLAG_HPDOFF = 4 /* some have *very* broken connector tables */
 };
 
 struct rhdCard {
@@ -50,6 +51,9 @@ struct rhdCard {
     enum rhdCardFlag flags;
 
     struct rhdConnectorInfo ConnectorInfo[RHD_CONNECTORS_MAX];
+#ifdef ATOM_BIOS
+    enum atomDevice DeviceInfo[RHD_CONNECTORS_MAX][MAX_OUTPUTS_PER_CONNECTOR];
+#endif
 };
 
 void RhdPrintConnectorInfo(int scrnIndex, struct rhdConnectorInfo *cp);
