@@ -27,15 +27,23 @@
 # include "config.h"
 #endif
 
+#define SEGV_ON_ASSERT 1		/* Define to 1 if you want backtraces on ASSERT() */
+
 #include "xf86.h"
+
 #if HAVE_XF86_ANSIC_H
 # include "xf86_ansic.h"
 #else
 # include <sys/types.h>
 # include <unistd.h>
-# include <signal.h>
 # include <string.h>
 # include <stdio.h>
+# include <stdlib.h>
+#endif
+
+#if SEGV_ON_ASSERT
+# include <sys/types.h>
+# include <signal.h>
 #endif
 
 #include "rhd.h"
@@ -199,18 +207,15 @@ RhdAppendString(char *s1, const char *s2)
     }
 }
 
-extern void xf86abort(void) NORETURN;
 void RhdAssertFailed(const char *str,
 		     const char *file, int line, const char *func)
 {
     ErrorF("%s:%d: %s: Assertion '%s' failed.\n", file, line, func, str);
 
-#if !HAVE_XF86_ANSIC_H		/* Set to 1 to get backtraces */
+#if SEGV_ON_ASSERT
     kill(getpid(), SIGSEGV);
-    xf86abort();	/* Not executed, but make gcc happy */
-#else
-    FatalError("Server aborting\n");
 #endif
+    FatalError ("Server aborting\n");
 }
 
 void RhdAssertFailedFormat(const char *str,
@@ -224,10 +229,8 @@ void RhdAssertFailedFormat(const char *str,
     va_end(args);
     ErrorF("\n");
 
-#if !(HAVE_XF86_ANSIC_H)		     /* Set to 1 to get backtraces */
+#if SEGV_ON_ASSERT
     kill(getpid(), SIGSEGV);
-    xf86abort();	/* Not executed, but make gcc happy */
-#else
-    FatalError("Server aborting\n");
 #endif
+    FatalError ("Server aborting\n");
 }
